from ._core._core import (
    confidence,
    conflicts,
    coverage,
    filter_labeled,
    is_labeled,
    matches,
    multipolar_to_monopolar,
    overlaps,
    summary,
    total_coverage,
)

__all__ = [
    "confidence",
    "conflicts",
    "coverage",
    "filter_labeled",
    "is_labeled",
    "matches",
    "multipolar_to_monopolar",
    "overlaps",
    "summary",
    "total_coverage",
]



import numpy as np
import pandas as pd


class LFS:
    """A container for managing labeling functions for data labeling.

    This class allows for the easy addition of labeling functions and
    their associated polarities.

    It also provides functionality to
    create a labeling matrix from a given DataFrame.

    Example
    -------
    >>> lfs = flippers.lfs.LFS()
    >>>
    >>> # Add a labeling function with polarity 1
    >>> @lfs.add(polarity=1)
    >>> def lf(df):
    >>>     return df["column_name"] > 0
    >>>
    >>> # Create a labeling matrix from a DataFrame
    >>> L = lfs.create_matrix(df)
    """

    def __init__(self) -> None:
        self.lfs = []
        self.polarities = []

    def add(self, polarity: int):
        """Adds a labeling function with its associated polarity.

        Parameters
        ----------
        polarity: The polarity associated with the labeling function.

        Example
        -------
        >>> @lfs.add(polarity=1)
        >>> def lf_happy(df):
        >>>     return df["text"].str.contains("happy")
        >>>
        >>> @lfs.add(polarity=0)
        >>> def lf_angry(df):
        >>>     return df["text"].str.contains("angry")
        """
        self.polarities.append(polarity)

        def decorator(lf):
            self.lfs.append(lf)
            return lf

        return decorator

    def create_matrix(self, df: pd.DataFrame):
        """Creates a labeling matrix from a given DataFrame.

        Each labeling function is applied to the DataFrame, and the
        results are stored in a matrix. The matrix is then returned as a
        DataFrame with the labeling function names as column names.

        Parameters
        ----------
        df : pd.DataFrame
            Input dataframe.

        Returns
        -------
        pd.DataFrame:
            Labeling matrix.

        Example
        -------
        >>> L_train = lfs.create_matrix(df_train)
        """
        L = np.zeros((len(df), len(self.lfs)))
        for i, lf in enumerate(self.lfs):
            L[:, i] = lf(df)
        column_names = [lf.__name__ for lf in self.lfs]

        L = pd.DataFrame(L, columns=column_names, index=df.index)
        return L



"""Implements a VAE to learn a latent parameter generating the weak label."""

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset
from tqdm import trange

from flippers._typing import ListLike, MatrixLike
from flippers.models._base import _BaseModel


class WeakLabelVAE(nn.Module, _BaseModel):
    """A label model implementation for weak supervision based on a VAE."""

    def __init__(
        self,
        polarities: ListLike,
        class_balances: ListLike = [],
        latent_dim: int = 4,
    ):
        """Initializes a SnorkelModel instance with the given configuration
        options.

        Parameters
        ----------
        polarities:
            List that maps weak labels to polarities, size n_weak.

        class_balances: ListLike, optional
            List specifying class balance prior for each possible class, size n_classes.

            Defaults to balanced classes prior.

        Example
        -------
        >>> polarities = [1, 0, 1, 1]
        >>> class_balances = [0.7, 0.3]
        >>> snorkel_model = SnorkelModel(polarities, cardinality, class_balances)
        """
        nn.Module.__init__(self)
        self.cardinality = 2
        _BaseModel.__init__(self, polarities, self.cardinality)

        self.Polarities = nn.Parameter(torch.Tensor(polarities), requires_grad=False)
        self.n_weak = len(polarities)
        self.latent_dim = latent_dim if latent_dim else self.n_weak // 2

        if not class_balances:
            class_balances = 1 / self.cardinality * np.ones(int(self.cardinality))
        self.class_balances = class_balances
        self.Balances = nn.Parameter(torch.Tensor(class_balances), requires_grad=False)

        self.encoder = nn.Sequential(
            nn.Linear(self.n_weak, self.n_weak),
            nn.ReLU(),
            nn.Linear(self.n_weak, 2 * self.latent_dim + 1),
        )

        self.decoder_mu = nn.Sequential(
            nn.Linear(self.latent_dim + 1, self.n_weak),
            nn.ReLU(),
            nn.Linear(self.n_weak, self.n_weak),
            nn.Sigmoid(),
        )
        self.loss_history = []
        self.mu_bar = nn.Parameter(torch.zeros(self.n_weak), requires_grad=True)

    @property
    def device(self):
        return next(self.parameters()).device

    def _convert_L(self, L: MatrixLike) -> torch.Tensor:
        """Convert input L to binary tensor."""
        L = np.array(L)
        L = L > 0.5
        L = torch.tensor(L).to(torch.float)
        L = L.to(self.device)
        return L

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std, device=self.device)
        return mu + eps * std

    def reparameterize_bool(self, p, eps=1e-8):
        U = torch.rand_like(p, device=self.device)
        gumbel_noise = -torch.log(-torch.log(U + eps) + eps)
        noisy_logit = p + gumbel_noise
        return torch.sigmoid(noisy_logit)

    def forward(self, L: torch.Tensor):
        mu_z, logvar_z, p = self.encoder(L).split(
            (self.latent_dim, self.latent_dim, 1), dim=1
        )
        z = self.reparameterize(mu_z, logvar_z)
        y = self.reparameterize_bool(p)

        latents = torch.cat((z, y), dim=1)

        mu = self.decoder_mu(latents)
        weak_reconstructed = (
            self.Polarities * y + (1 - self.Polarities) * (1 - y)
        ) * mu + self.mu_bar
        weak_reconstructed = weak_reconstructed.clamp(1e-8, 1 - 1e-8)

        return weak_reconstructed, mu_z, logvar_z, p

    def predict_proba(self, L):
        L = self._convert_L(L)

        with torch.no_grad():
            mu_z, logvar_z, p = self.encoder(L).split(
                (self.latent_dim, self.latent_dim, 1), dim=1
            )

            p = torch.sigmoid(p)

            p = torch.cat((1 - p, p), dim=1)

        p = p.detach().cpu().numpy()

        return p

    def loss(self, L, outputs, kld_weight, nudge, capacity):
        L_reconstructed, mu_z, logvar_z, p = outputs

        loss_L = (
            nn.functional.binary_cross_entropy(L_reconstructed, L, reduction="none")
            .mean(dim=1)
            .sum()
        )

        kl_z = -0.5 * (1 + logvar_z - mu_z.pow(2) - logvar_z.exp()).mean(dim=1).sum()

        # Nudge from 0.5
        # Force mean to be class balance
        p = torch.sigmoid(p)
        p_mean = p.mean(dim=0)
        y = self.class_balances[1]
        kl_p = y * torch.log(y / p_mean) + (1 - y) * torch.log((1 - y) / (1 - p_mean))
        kl_p = (kl_p - capacity).relu()
        kl_p = kl_p * L.shape[0]

        return loss_L + kld_weight * kl_z + nudge * kl_p

    def _get_dataloader(self, L, batch_size):
        class CustomDataset(Dataset):
            def __init__(self, L):
                self.L = L

            def __len__(self):
                return len(self.L)

            def __getitem__(self, idx):
                return self.L[idx]

        dataloader = DataLoader(
            CustomDataset(L), batch_size=batch_size, shuffle=True, drop_last=True
        )

        return dataloader

    def fit(
        self,
        L: MatrixLike,
        learning_rate: float = 5e-3,
        num_batches: int = 5000,
        batch_size: int = 32,
        weight_decay: float = 1e-3,
        kld_weight: float = 1,
        nudge: float = 1,
        capacity=1e-1,
        verbose: bool = True,
        **_,
    ) -> None:
        """Train the Snorkel model on the given weak label matrix L.

        Parameters
        ----------
        L : MatrixLike
            Weak Label matrix of shape (num_samples, n_weak)
        learning_rate : float, optional, default: 1e-3
            Learning rate for the optimizer.
        num_batches : int, optional, default: 5000
            Number of batches to train the model
        weight_decay : float, optional, default: 0
            Weight decay (L2 penalty) for the optimizer
        kld_weight : float, optional, default: 0
            Weight of class blance loss term.

            This term penalizes the model for predicting a class on the train set
            differently to its specified balance
        device : str, optional, default: "cpu"
            Device to use for training, either "cpu" or "cuda"
        verbose : bool, optional, default: False
            If True, displays training progress

        Returns
        -------
        None

        Example
        -------
        >>> L = [
        ...     [1, 0, 1, 1],
        ...     [0, 1, 0, 1],
        ...     [1, 0, 1, 0]
        ... ]
        >>> model.fit(
        ...     L,
        ...     learning_rate=1e-2,
        ...     num_batches=3000,
        ...     prec_init=0.7,
        ...     k=5e-3,
        ...     device="cpu",
        ...     verbose=True
        ... )
        """

        # Convert L to binary tensor
        L = self._convert_L(L)

        optimizer = optim.AdamW(
            self.parameters(), lr=learning_rate, weight_decay=weight_decay
        )
        num_epochs = num_batches * batch_size // L.shape[0] + 1

        epoch_range = trange(num_epochs) if verbose else range(num_epochs)

        dataloader = self._get_dataloader(L, batch_size)

        for epoch in epoch_range:
            epoch_loss = 0
            for batch_L in dataloader:
                # Zero the gradients
                optimizer.zero_grad()

                batch_L = batch_L.to(self.device)

                outputs = self(batch_L)
                loss = self.loss(
                    batch_L,
                    outputs,
                    kld_weight=kld_weight,
                    nudge=nudge,
                    capacity=capacity,
                )

                loss.backward()
                optimizer.step()

                epoch_loss += loss.item()

            if verbose:
                epoch_loss /= len(dataloader)
                epoch_range.set_description(f"Epoch [{epoch+1}/{num_epochs}]")
                description = {}
                # Log loss_history
                self.loss_history.append(epoch_loss)
                description["Loss"] = f"{epoch_loss:.1f}"
                epoch_range.set_postfix(description)



__title__ = "flippers"
__summary__ = "A Python library for weak supervision."
__url__ = "https://github.com/liamtoran/flippers"
__version__ = "0.0.1-beta4"
__author__ = "Liam Toran"
__author_email__ = "liam.toran@gmail.com"
__license__ = "Apache-2.0"



import warnings
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd

from .._typing import ListLike, MatrixLike


class _WeakLabelInfo:
    """Collects information about the weak labelers."""

    def __init__(self, polarities: ListLike, cardinality: int = 0):
        """
        Parameters
        ----------
        polarities
            List that maps weak labels to polarities.
        cardinality: int, optional
            Number of classes

            If not specified, it will be inferred from the maximum value in polarities.

        """
        # Set polarities
        if isinstance(polarities, list):
            self.polarities = np.array(polarities)
        elif isinstance(polarities, np.ndarray):
            self.polarities = polarities
        else:
            ValueError("Input polaritiy is not a list or NumPy array")

        # Set cardinality
        if not cardinality:
            # Infer the number of categories from the maximum value of polarities
            cardinality = max(self.polarities) + 1
        self.cardinality = cardinality

        # Validate inputs work
        self.__validate_init__()

        self.n_weak = len(self.polarities)
        self.polarities_matrix = self._get_polarity_matrix()
        self.overlap_matrix = self._get_overlap_matrix()

    def __validate_init__(self):
        """Assert whether the input data is valid."""
        # Check more than one class
        assert self.cardinality > 1

        # Check no polarities outside of cardinality
        assert self.polarities.max() <= self.cardinality - 1
        assert self.polarities.min() >= 0

    def _get_polarity_matrix(self):
        """Convert polarities to a binary matrix.

        Returns
        -------
            Binary matrix of shape (n_weak, cardinality) where:

            mask[i,j] = polarity[i] == j
        """
        # Use broadcasting to create a boolean mask of the elements that
        # match each possible value of polarities
        mask = self.polarities[:, np.newaxis] == np.arange(self.cardinality)

        return mask

    def _get_overlap_matrix(self):
        polarity_overlap = self.polarities_matrix @ self.polarities_matrix.T
        return polarity_overlap


def multipolar_to_monopolar(
    L: MatrixLike, polarities_mapping: Dict[str, List[int]] = {}
) -> Tuple[pd.DataFrame, np.ndarray, Dict[str, List[int]]]:
    """Convert a pandas DataFrame of weak labels in multipolar representation
    to monopolar representation.

    Parameters
    ----------
    L : pd.DataFrame
        The input DataFrame of weak labels.

        Each column represents a weak labeler and each row represents a data point.
    polarities_mapping :
        A dictionary specifying the possible polarities for each weak labeler.

        The keys are the column names of L and the values are lists of integers
        representing the possible polarities.

        If not specified, the function
        will attempt to infer the polarities by examining the unique values in
        each column of L.

    Returns
    -------
    L_monopolar, polarities, polarities_mapping
        A tuple containing the following elements:

        - A pandas DataFrame of monopolar weak labels.

        - A 1D numpy array containing the polarities

        - A dictionary specifying the original polarities for each weak labeler.

        The keys are the column names of the input and the values are lists of integers
        representing the possible polarities.

    Example
    -------
    Its always better to use an hand written ``polairites_mapping``.
    ``polarities_mapping`` lists possible polarities each labeling function can have.

    >>> multipolar = pd.DataFrame([[-1, 1, 2], [0, -1, 0], [-1, -1, 2]])
    >>> polarities_mapping = {'0': [0], '1': [1], '2': [0, 2]}
    >>> L, polarities, _ = flippers.multipolar_to_monopolar(
                                                multipolar, polarities_mapping
                                                )
    >>> L
       0  1  2__0  2__2
    0  0  1     0     1
    1  1  0     1     0
    2  0  0     0     1

    If you dont want to create the mapping, the function can infer one.
    This is potentially breaking if multipolar does not contain all possible outputs.

    >>> L, polarities, polarities_mapping = flippers.multipolar_to_monopolar(multipolar)
    >>> L
       0  1  2__0  2__2
    0  0  1     0     1
    1  1  0     1     0
    2  0  0     0     1
    >>> polarities # output L polarities
    array([0, 1, 0, 2], dtype=int64)
    >>> polarities_mapping # input multipolar polarities
    {'0': [0], '1': [1], '2': [0, 2]}
    """
    L = pd.DataFrame(L)

    L.columns = L.columns.astype(str)
    new_columns = {}
    polarities = []

    # Create polarity mapping if not furnished
    if not polarities_mapping:
        warnings.warn(
            (
                "Polarity mapping if not furnished."
                "\nMake sure L captures all possible values of each weak labelers."
            )
        )
        polarities_mapping = {}

        for col in L.columns:
            values = L[col].unique()
            values = list(values[values >= 0])
            values.sort()

            if len(values) == 0:
                warnings.warn(
                    f"Column {col} is never labeled, assuming its polarity is 0"
                )
                values = [0]

            polarities_mapping[str(col)] = values

            for polarity in values:
                new_col = (L[col] == polarity).astype(int)
                new_columns[col + (len(values) > 1) * f"__{polarity}"] = new_col
                polarities.append(polarity)

    # Use polarity mapping
    # This is more robust but requires using a polarities mapping dict
    else:
        for col, values in polarities_mapping.items():
            for polarity in values:
                polarities.append(polarity)
                new_col = (L[col] == polarity).astype(int)
                new_columns[col + (len(values) > 1) * f"__{polarity}"] = new_col

    new_L = pd.DataFrame.from_dict(new_columns)
    polarities = np.array(polarities)

    return new_L, polarities, polarities_mapping


def is_labeled(L: pd.DataFrame) -> pd.Series:
    """Check if any labels exist in the given label matrix L.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).

    Returns
    -------
    Series of size n_samples indicating whether a sample is labeled or not.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.is_labeled(L)
    0     True
    1     True
    2    False
    3     True
    dtype: bool
    """
    L = pd.DataFrame(L)
    return L.sum(axis=1) > 0


def total_coverage(L: pd.DataFrame) -> float:
    """Calculate the total proportion of labeled samples in the given label
    matrix.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).

    Returns
    -------
    float
        Total coverage, ranging from 0 to 1, indicating the proportion of
        labeled samples in the label matrix.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.total_coverage(L)
    0.75
    """
    L = pd.DataFrame(L)
    return is_labeled(L).mean()


def filter_labeled(L: pd.DataFrame) -> pd.DataFrame:
    """Filter out unlabeled samples from the given label matrix.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).

    Returns
    -------
    Returns a filtered label matrix of shape (n_labeled_samples, n_weak).
    Sliced on the condition that the first one is labeled.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.filter_labeled(L)
       0  1  2
    0  0  1  0
    2  1  0  1
    """
    L = pd.DataFrame(L)
    labeled = is_labeled(L)
    return L.loc[labeled]


def coverage(L: pd.DataFrame) -> pd.Series:
    """Calculate the average of samples labeled per weak labeler.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).

    Returns
    -------
    Series of size n_weak with average of samples labeled per weak labeler.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.coverage(L)
    0    0.5
    1    0.5
    2    0.5
    dtype: float64
    """
    L = pd.DataFrame(L)
    return (L > 0).mean()


def confidence(L: pd.DataFrame) -> pd.Series:
    """Calculate the average confidence level per weak labeler.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak)1.

    Returns
    -------
    Series of size n_weak with average confidence level per weak labeler.

    Example
    -------
    >>> L = pd.DataFrame([[0, .1, 0], [1, 0, .5], [0, 0, 0], [.7, .1, .2]])
    >>> flippers.confidence(L)
    0    0.85
    1    0.10
    2    0.35
    dtype: float64
    """
    L = pd.DataFrame(L)
    return L[L > 0].mean()


def _overlaps(
    L: pd.DataFrame, polarities: ListLike = [], sign: str = "overlaps"
) -> pd.Series:
    """Calculate the number of overlaps for each sample and weak label.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).
    polarities : Union[list, np.ndarray]
        Array or list of size n_weak containing the polarity of each weak label.

        Optional if sign="overlaps".
    sign : str, optional (default="all", options "overlaps", "matches", "conflicts")
        String specifying which overlaps to include. Valid values are:

        - "overlaps" (default) to include both positive and negative overlaps,

        - "matches" to include matching overlaps only,

        - "conflicts" to include negative overlaps only.

    Returns
    -------
    pd.Series
        Series of length n_weak indicating the fraction of
        annotated samples with other annotations for each LF.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> # All overlaps
    >>> flippers._overlaps(L)
    0    0.50
    1    0.25
    2    0.50
    dtype: float64
    >>> # Only overlaps with matching assigned label
    >>> flippers._overlaps(L, polarities, sign="matches")
    0    0.00
    1    0.25
    2    0.25
    dtype: float64
    >>> # Only overlaps with conflicting assigned label
    >>> flippers._overlaps(L, polarities, sign="conflicts")
    0    0.50
    1    0.25
    2    0.50
    dtype: float64
    """
    names = pd.DataFrame(L).columns
    L = np.array(L)

    all_overlap = L.sum(axis=1).reshape(-1, 1) * L - L
    if sign == "overlaps":
        overlap = all_overlap
    else:
        mask = _WeakLabelInfo(polarities).overlap_matrix
        match_overlap = L @ mask * L - L
    if sign == "matches":
        overlap = match_overlap
    elif sign == "conflicts":
        overlap = all_overlap - match_overlap

    overlap = overlap > 0
    overlap = overlap.sum(axis=0)
    overlap = overlap / len(L)

    overlap = pd.Series(overlap, index=names)
    return overlap  # type: ignore


def overlaps(
    L: pd.DataFrame,
) -> pd.Series:
    """Calculate the number of fraction of labeled samples labeled by other
    labeling functions for each labeling function.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).

    Returns
    -------
    pd.Series
        Series of length n_weak indicating the fraction of
        annotated samples with other annotations for each LF.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.overlap(L)
    0    0.50
    1    0.25
    2    0.50
    dtype: float64
    """
    overlap = _overlaps(L)
    return overlap  # type: ignore


def conflicts(L: pd.DataFrame, polarities: ListLike) -> pd.Series:
    """Calculate the number of fraction of labeled samples labeled differently
    by other labeling functions for each labeling function.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).
    polarities : Union[list, np.ndarray]
        Array or list of size n_weak containing the polarity of each weak label.

    Returns
    -------
    pd.Series
        Series of length n_weak indicating the fraction of
        annotated samples with conflicting annotations for each LF.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.conflicts(L, polarities)
    0    0.50
    1    0.25
    2    0.50
    dtype: float64
    """
    overlap = _overlaps(L, polarities, sign="conflicts")
    return overlap  # type: ignore


def matches(L: pd.DataFrame, polarities: ListLike) -> pd.Series:
    """Calculate the number of fraction of labeled samples labeled similarly by
    other labeling functions for each labeling function.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).
    polarities : Union[list, np.ndarray]
        Array or list of size n_weak containing the polarity of each weak label.

    Returns
    -------
    pd.Series
        Series of length n_weak indicating the fraction of
        annotated samples with matching annotations for each LF.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> flippers.matches(L, polarities)
    0    0.00
    1    0.25
    2    0.25
    dtype: float64
    """
    overlap = _overlaps(L, polarities, sign="matches")
    return overlap  # type: ignore


def summary(
    L: MatrixLike, polarities: ListLike, digits: int = 3, normalize: int = False
) -> pd.DataFrame:
    """Calculate summary statistics for the given weak label matrix and
    polarities.

    Parameters
    ----------
    L : pd.DataFrame
        Weak label DataFrame of shape (n_samples, n_weak).
    polarities:
        1D array or list of size n_weak containing the polarities of each weak label.
    digits:
        Number of digits to round the output statistics to. Default 3.
    normalize:
        When True, shows overlaps/matches/conflicts as a ratio of coverage.

    Returns
    -------
    DataFrame of shape (n_weak, n_summaries) containing the following columns
        - "polarity": The polarity of each weak label.
        - "coverage": The average ratio of samples that are assigned each weak label.
        - "confidence": The average confidence level of the assigned weak labels.
        - "overlaps": The ratio of assigned labels that have overlapping labels.
        - "matches": The ratio of assigned labels that have other matching labels.
        - "conflicts": The ratio of assigned labels that have conflicting labels.

    Example
    -------
    >>> L = pd.DataFrame([[0, 1, 0], [1, 0, 1], [0, 0, 0], [1, 1, 1]])
    >>> polarities = [0, 1, 1]
    >>> flippers.summary(L, polarities)
        polarity  coverage  confidence  overlaps  matches  conflicts
    0         0       0.5         1.0      0.50     0.00        0.50
    1         1       0.5         1.0      0.25     0.25        0.25
    2         1       0.5         1.0      0.50     0.25        0.50
    >>> flippers.summary(L, polarities, normalize=True)
        polarity  coverage  confidence  overlaps  matches  conflicts
    0         0       0.5         1.0       1.0      0.0         1.0
    1         1       0.5         1.0       0.5      0.5         0.5
    2         1       0.5         1.0       1.0      0.5         1.0
    """
    L = pd.DataFrame(L)

    descriptions = pd.DataFrame(index=L.columns)
    descriptions["polarity"] = polarities
    descriptions["coverage"] = coverage(L)
    descriptions["confidence"] = confidence(L)
    descriptions["overlaps"] = overlaps(L)
    descriptions["matches"] = matches(L, polarities)
    descriptions["conflicts"] = conflicts(L, polarities)
    if normalize:
        descriptions["overlaps"] /= descriptions["coverage"]
        descriptions["matches"] /= descriptions["coverage"]
        descriptions["conflicts"] /= descriptions["coverage"]
    descriptions = descriptions.round(digits)

    return descriptions


assert summary(
    np.array(
        [
            [1, 1, 0, 1],
            [0, 0, 1, 0],
            [0, 1, 0, 1],
            [0, 0, 0, 1],
            [0, 0, 0, 1],
            [0, 0, 0, 1],
            [1, 1, 0, 1],
            [0, 0, 1, 1],
            [0, 1, 0, 1],
            [0, 0, 0, 1],
            [0, 0, 1, 0],
            [0, 1, 0, 1],
            [1, 1, 0, 0],
            [0, 1, 0, 0],
        ]
    ),
    [0, 1, 1, 0],
).to_dict() == {
    "polarity": {0: 0, 1: 1, 2: 1, 3: 0},
    "coverage": {0: 0.214, 1: 0.5, 2: 0.214, 3: 0.714},
    "confidence": {0: 1.0, 1: 1.0, 2: 1.0, 3: 1.0},
    "overlaps": {0: 0.214, 1: 0.429, 2: 0.071, 3: 0.429},
    "matches": {0: 0.143, 1: 0.0, 2: 0.0, 3: 0.143},
    "conflicts": {0: 0.214, 1: 0.429, 2: 0.071, 3: 0.429},
}



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.calibration import calibration_curve

import flippers

binary_datasets = [
    "imdb",
    "yelp",
    "youtube",
    "sms",
    "spouse",
    "cdr",
    "commercial",
    "tennis",
    "basketball",
    "census",
]

dataset_to_metric = {
    "imdb": "Accuracy",
    "yelp": "Accuracy",
    "youtube": "Accuracy",
    "sms": "F1",
    "agnews": "Accuracy",
    "trec": "Accuracy",
    "spouse": "F1",
    "cdr": "F1",
    "semeval": "Accuracy",
    "chemprot": "Accuracy",
    "commercial": "F1",
    "tennis": "F1",
    "basketball": "F1",
    "census": "F1",
}

datasets = list(dataset_to_metric.keys())


# This helper loads any dataset in the wrench benchmarks
# and creates monopolar label matrices from their multipolar data
def load_wrench_dataset(dataset):
    dataset = "datasets/" + f"{dataset}" + "/{}.json"
    train = pd.read_json(dataset.format("train")).T
    dev = pd.read_json(dataset.format("valid")).T
    test = pd.read_json(dataset.format("test")).T

    L_train, polarities, polarities_mapping = flippers.multipolar_to_monopolar(
        train["weak_labels"].apply(pd.Series)
    )
    L_dev, _, _ = flippers.multipolar_to_monopolar(
        dev["weak_labels"].apply(pd.Series), polarities_mapping
    )
    L_test, _, _ = flippers.multipolar_to_monopolar(
        test["weak_labels"].apply(pd.Series), polarities_mapping
    )

    return (train, dev, test), (L_train, L_dev, L_test), polarities


class MetricsUtil:
    def __init__(self, y_true, L=None):
        self.L = L
        self.y_true = y_true
        self.n_classes = len(np.unique(y_true))
        self.metrics = {}

    def score(
        self,
        model=None,
        name="",
        y_pred=None,
        plots=True,
        fill_proba=False,
        predict_proba_args={},
    ):
        if y_pred is None:
            y_pred = model.predict_proba(self.L, **predict_proba_args)

        if fill_proba:

            def fill_proba(proba):
                proba = proba.copy()
                proba[self.L.sum(axis=1) == 0] = 0
                return proba

            y_pred = fill_proba(y_pred)
        if self.n_classes == 2:
            y_pred = y_pred[:, 1]
            AP = metrics.average_precision_score(self.y_true, y_pred)
            F1 = metrics.f1_score(self.y_true, y_pred.round())
            AUC = metrics.roc_auc_score(self.y_true, y_pred)
            Accuracy = metrics.accuracy_score(self.y_true, y_pred.round())
            Balanced_Accuracy = metrics.balanced_accuracy_score(
                self.y_true, y_pred.round()
            )
            M = {
                "F1": F1,
                "Average_Precision": AP,
                "AUC": AUC,
                "Accuracy": Accuracy,
                "Balanced_Accuracy": Balanced_Accuracy,
            }
        else:
            M = {}
            # for average in ["macro"]:
            #     precision, recall, fscore, _ = precision_recall_fscore_support(
            #         self.y_true, y_pred.argmax(axis=1), average=average
            #     )
            #     M[f"Precision_{average}"] = precision
            #     M[f"Recall_{average}"] = recall
            #     M[f"F1_{average}"] = fscore
            M["Accuracy"] = metrics.accuracy_score(self.y_true, y_pred.argmax(axis=1))
            M["Balanced_Accuracy"] = metrics.balanced_accuracy_score(
                self.y_true, y_pred.argmax(axis=1)
            )

        for key in M:
            M[key] = round(M[key], 3)

        if plots and self.n_classes == 2:
            fig, axs = plt.subplots(1, 2, figsize=(9, 4))
            plt.title(name)

            # Plot the distribution of y_pred grouped by ground truth
            pd.DataFrame({"y_pred": y_pred, "y_true": self.y_true}).boxplot(
                by="y_true", ax=axs[0]
            )
            axs[0].set_title("Predicted probabilities grouped by ground truth")
            axs[0].set_ylim([0, 1])
            axs[0].get_figure().suptitle("")
            axs[1].set_ylabel("y_pred")

            # Plot the distribution of y_pred
            y_pred = pd.Series(y_pred)
            y_pred.plot.hist(
                ax=axs[1],
                bins=21,
                weights=np.ones_like(y_pred.index) / len(y_pred.index),
            )
            axs[1].set_title("Predicted probabilties")
            #  axs[1].set_ylim([0, 1])
            axs[1].scatter(
                [self.y_true.mean()],
                [0.1],
                color="g",
                marker="o",
                s=100,
                alpha=0.5,
                label="Mean of y_true",
            )

            # Add a cross indicating the mean of y_pred
            mean_y_pred = pd.Series(y_pred).mean()
            axs[1].scatter(
                [mean_y_pred],
                [0.1],
                marker="+",
                color="r",
                s=200,
                label="Mean of y_pred",
            )

            ## Calibration curve
            # Compute the calibration curve
            fraction_of_positives, mean_predicted_value = calibration_curve(
                self.y_true, y_pred, n_bins=20
            )

            # Plot the calibration curve
            axs[1].set_title("Calibration Curve")
            axs[1].plot(
                mean_predicted_value,
                fraction_of_positives,
                "s-",
                label="Fraction of positives",
                alpha=0.5,
            )
            axs[1].plot([0, 1], [0, 1], "--", color="gray")
            axs[1].set_xlabel("Predicted Value")
            # axs[1].set_ylabel("Fraction of Positives")
            axs[1].set_xlabel("y_pred")
            axs[1].legend()

            plt.tight_layout()
            plt.show()

        if name:
            self.metrics[name] = M

        return M



"""Implements the snorkel library label model."""


import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from tqdm import trange

from .._typing import ListLike, MatrixLike
from ._base import _BaseModel


class SnorkelModel(nn.Module, _BaseModel):
    """A label model implementation for weak supervision based on a generative
    approach.

    This implementation is based on the Snorkel library's label model.

    Like its snorkel library counterpart assumes that the labeling
    functions are independent conditionally to Y, similar to a naive
    Bayes assumption.

    However, good results can also be observed in practice for
    correlated LFs.

    See the following link[] for more information on how to use this
    model and a comparison with the Snorkel library's implementation.

    Example
    -------

    >>> from flippers.models import SnorkelModel
    >>>
    >>> # Create a SnorkelModel instance
    >>> label_model = SnorkelModel(polarities, cardinality, class_balances)
    >>> # Train the model
    >>> label_model.fit(L_train)
    >>> # Generate labels
    >>> y_pred_proba = label_model.predict_proba(L) # shape: (len(L), n_classes)
    >>> y_pred = label_model.predict(L) # shape: (len(L),)
    """

    def __init__(
        self,
        polarities: ListLike,
        cardinality: int = 0,
        class_balances: ListLike = [],
    ):
        """Initializes a SnorkelModel instance with the given configuration
        options.

        Parameters
        ----------
        polarities:
            List that maps weak labels to polarities, size n_weak.
        cardinality: int, optional
            Number of possible label values.

            If unspecified, it will be inferred from the maximum value in polarities.
        class_balances: ListLike, optional
            List specifying class balance prior for each possible class, size n_classes.

            Defaults to balanced classes prior.

        Example
        -------
        >>> polarities = [1, 0, 1, 1]
        >>> cardinality = 2
        >>> class_balances = [0.7, 0.3]
        >>> label_model = SnorkelModel(polarities, cardinality, class_balances)
        >>> # Change device
        >>> label_model.to("cuda")
        """
        nn.Module.__init__(self)
        _BaseModel.__init__(self, polarities, cardinality)

        self.Polarities = nn.Parameter(
            torch.Tensor(self.polarities_matrix), requires_grad=False
        )

        if not class_balances:
            class_balances = 1 / self.cardinality * np.ones(int(self.cardinality))
        self.class_balances = class_balances
        self.Balances = nn.Parameter(torch.Tensor(class_balances), requires_grad=False)
        self.loss_history = []

    @property
    def device(self):
        return next(self.parameters()).device

    def _convert_L(self, L: MatrixLike) -> torch.Tensor:
        """Convert input L to binary tensor."""
        L = np.array(L)
        L = L > 0.5
        L = torch.tensor(L).to(device=self.device, dtype=torch.float32)
        return L

    def fit(
        self,
        L: MatrixLike,
        learning_rate: float = 1e-3,
        num_epochs: int = 50,
        prec_init: float = 0.5,
        k: float = 0,
        verbose: bool = False,
        **_,
    ) -> None:
        """Train the Snorkel model on the given weak label matrix L.

        Parameters
        ----------
        L : MatrixLike
            Weak Label matrix of shape (num_samples, n_weak)
        learning_rate : float, optional, default: 1e-3
            Learning rate for the optimizer.
        num_epochs : int, optional, default: 50
            Number of epochs to train the model
        prec_init : float, optional, default: 0.5
            Initial value for precision

            Can be of shape (n_weak) to set precision for each LF.
        k : float, optional, default: 0
            Weight of class blance loss term.

            This term penalizes the model for predicting a class on the train set
            differently to its specified balance
        verbose : bool, optional, default: False
            When True, displays training progress using tqdm.

        Returns
        -------
        None

        Example
        -------
        >>> L = [
        ...     [1, 0, 1, 1],
        ...     [0, 1, 0, 1],
        ...     [1, 0, 1, 0]
        ... ]
        >>> label_model.fit(
        ...     L,
        ...     learning_rate=1e-2,
        ...     num_epochs=10,
        ...     prec_init=0.7,
        ...     k=5e-3,
        ...     verbose=True
        ... )
        """

        self.num_epochs = num_epochs
        self.learning_rate = learning_rate
        eps = 1e-6

        # Convert L to binary tensor
        L = self._convert_L(L)

        # Gram matrix of L gives overlaps
        Overlaps = L.T @ L / len(L)
        Coverage = torch.diag(Overlaps)  # Overlaps with itself = coverage
        Overlaps, Coverage = Overlaps.to(self.device), Coverage.to(self.device)

        # mu(i, j) = P(L_i = 1 | Y = j)
        # Bayes theorem
        if not hasattr(self, "prec_init"):
            self.prec_init = prec_init
            self.Prec_init = torch.ones(self.n_weak).to(self.device) * prec_init
            self.mu_init = (
                self.Polarities
                * (self.Prec_init * Coverage).view(-1, 1)
                / self.Balances
            ).float()
            self.mu = nn.Parameter(self.mu_init.clone())  # * np.random.random())

        # This will be changed when dealing with cliques the right way
        mask_overlap = torch.ones(self.n_weak, self.n_weak).to(self.device).int()
        mask_overlap = mask_overlap - torch.diag(torch.diag(mask_overlap))
        mask_overlap = mask_overlap.bool()
        Overlaps = Overlaps[mask_overlap]

        optimizer = optim.Adam(
            [x for x in self.parameters() if x.requires_grad],
            lr=self.learning_rate,
        )

        def loss_fn(x, y):
            return torch.norm(x - y) ** 2

        if verbose:
            epoch_range = trange(self.num_epochs)
        else:
            epoch_range = range(self.num_epochs)

        for epoch in epoch_range:
            # Zero the gradients
            optimizer.zero_grad()
            self.mu.data = self.mu.clamp(eps, 1 - eps)

            # # Forward

            # Overlap reconstruction loss
            overlap_reconstructed = ((self.mu * self.Balances) @ self.mu.t())[
                mask_overlap
            ]
            loss_overlap = loss_fn(overlap_reconstructed, Overlaps)

            # Coverage reconstruction loss
            # coverage = mu @ class_balance by total law of probability
            loss_coverage = loss_fn(self.mu @ self.Balances, Coverage)

            loss = loss_overlap + loss_coverage
            if k > 0:
                loss_prediction = loss_fn(
                    self.Balances, self._predict_proba_tensor(L).mean(0)
                )

                loss = loss + k * loss_prediction

            # Backward
            loss.backward()

            # Optimizer Step
            optimizer.step()

            # Log loss_history
            self.loss_history.append(loss.item())

            if verbose and ((epoch % 50 == 0) or (epoch == self.num_epochs - 1)):
                epoch_range.set_description(
                    f"Epoch {epoch+1}/{self.num_epochs} | Loss {loss.item():.4f}"
                )

        # Clamp mu
        self.mu.data = self.mu.clamp(eps, 1 - eps)

        # Eval mode so predict_proba doesnt calculate gradients
        self.eval()

    def _predict_proba_tensor(
        self, L: torch.tensor, ignore_abstains: bool = False
    ) -> torch.tensor:
        # Computing trick :
        # exp(L @ log(mu))[i, k] = Product of mu[j, k] for j / L[i, j] = 1
        Log_likelihood = L @ self.mu.log()
        if not ignore_abstains:
            # Include likelihood of abstains
            Log_likelihood = Log_likelihood + ((1 - L) @ (1 - self.mu).log())

        # Update prior
        Proba = Log_likelihood.exp() * self.Balances
        Proba = Proba / Proba.sum(1).view(-1, 1)
        return Proba

    def predict_proba(self, L: MatrixLike, ignore_abstains: bool = False) -> np.ndarray:
        """Predicts the probabilities of the classes by updating the prior
        using the learned parameter mu as posteriors.

        Parameters
        ----------
        L : MatrixLike
            Weak Label matrix
        ignore_abstains : bool, optional
            Whether to ignore abstains in the prior update:

            $ When `False` (default), uses both votes and abstains. This is recommended
            and helps leverage information gained from
            knowing which labeling function abstained.

            $ When `True`, updates prior only using non abstained.

        Returns
        -------
        numpy.ndarray
            An array of predicted probabilities of shape (num_samples, num_classes).

        Example
        -------
        >>> L = [
        ...     [1, 0, 1, 1],
        ...     [0, 1, 0, 1],
        ...     [1, 0, 1, 0]
        ... ]
        >>> proba = label_model.predict_proba(L)
        >>> # proba.shape = (len(L), cardinality)
        """
        L = self._convert_L(L)

        with torch.no_grad():
            Proba = self._predict_proba_tensor(L, ignore_abstains)

        proba = Proba.detach().cpu().numpy()

        return proba



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from datetime import datetime

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
# Import the theme and set it as the default

project = "flippers"
copyright = f"2023 - {datetime.now().year}, Liam Toran"
author = "Liam Toran"
release = ""


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx_argparse_cli",
    "sphinx_copybutton",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
autosectionlabel_prefix_document = True
autosummary_imported_members = True
autosummary_generate = True
autosummary_recursive = True

autodoc_typehints = "description"
autoclass_content = "both"
autodoc_mock_imports = ["torch", "tqdm", "_base.py", "_core.py"]
autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "show-inheritance": True,
    "member-order": "bysource",
    "special-members": "*predict, *predict_proba, *save, *load",
    "exclude-members": "__weakref__",
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = [
    "custom.css",
]
html_sidebars = {
    "**": ["search-field.html", "sidebar-nav-bs.html", "sidebar-ethical-ads.html"]
}
html_theme_options = {
    "github_url": "https://github.com/liamtoran/flippers",
}



"""Groups basic generative models."""

import pickle
import warnings
from abc import ABC, abstractmethod

import numpy as np

from .._core import _WeakLabelInfo
from .._typing import ListLike, MatrixLike


class _BaseModel(_WeakLabelInfo, ABC):
    """Create a Model object."""

    def __init__(
        self,
        polarities: ListLike,
        cardinality: int = 0,
    ):
        """
        Parameters
        ----------
        polarities:
            List that maps weak labels to polarities, size n_weak.
        cardinality:
            Number of possible label values.

            If unspecified, it will be inferred from the maximum value in polarities.

        Example
        -------
        >>> polarities = [1, 0, 1, 1]
        >>> cardinality = 2
        >>> model = ModelClass(polarities, cardinality)
        """
        ABC.__init__(self)
        _WeakLabelInfo.__init__(self, polarities, cardinality)

    def save(self, filepath):
        """Save the model to a file.

        Parameters
        ----------
        filepath : str
            Path to the file where the model will be saved.

        Example
        -------
        >>> model.save("label_model.pkl")
        """
        with open(filepath, "wb") as file:
            pickle.dump(self, file)

    @classmethod
    def load(cls, filepath):
        """Load a saved model from a file.

        Parameters
        ----------
        filepath : str
            Path to the file containing the saved model.

        Returns
        -------
        The loaded model object.

        Example
        -------
        >>> model = ModelClass.load("label_model.pkl")
        """
        with open(filepath, "rb") as file:
            model = pickle.load(file)
        return model

    @abstractmethod
    def predict_proba(self, L: MatrixLike) -> np.ndarray:
        """Predict probabilites for the given weak label matrix.

        Parameters
        ----------
        L:
            Weak label dataframe.

            Shape: (n_samples, n_weak)

        Returns
        -------
            Array of predicted probabilities of shape (len(L), cardinality)


        Example
        -------
        >>> L = [[1, 0, 1, 2], [0, 1, 0, 0]]
        >>> proba = base_model.predict_proba(L)
        >>> # proba.shape = (len(L), cardinality)
        """
        pass

    def predict(self, L: MatrixLike, strategy: str = "majority") -> np.ndarray:
        """Predict labels for the given weak label matrix using the specified
        strategy.

        Parameters
        ----------
        L
            Weak label dataframe.

            Shape: (n_samples, n_weak)
        strategy
            Prediction strategy to use. Supported values: majority, probability.

            Controls how labels are predicted from the predicted probabilites.

            - majority: Predict the label with the highest number of votes.

            - probability: Predict label j with probability proba[i, j].

              This can be useful to enforce specific class_balances in the predictions.

            Default is "majority".

            If there are no votes for a sample, will predict -1.

        Returns
        -------
            1-D array of predicted labels of size n_samples

        Example
        -------
        >>> L = [[1, 0, 1, 2], [0, 1, 0, 0]]
        >>> predictions = base_model.predict(L)
        >>> # predictions.shape = (len(L),)
        """
        proba = self.predict_proba(L)
        unlabeled = proba.sum(axis=1) == 0
        if strategy == "majority":
            # Predict -1 if no votes were cast for all categories
            # Else predict the majority
            predictions = np.where(unlabeled, -1, proba.argmax(axis=1))
        elif strategy == "probability":
            # Predict -1 if no votes were cast for all categories
            # Else use probability matching / Thompson sampling
            filled_proba = proba.copy()
            filled_proba[unlabeled] = 1 / self.cardinality
            predictions = np.apply_along_axis(
                lambda x: np.random.choice(self.cardinality, p=x),
                axis=1,
                arr=filled_proba,
            )
            predictions = np.where(unlabeled, -1, predictions)
        else:
            raise ValueError('strategy should be "majority" or "probability"')

        return predictions

    def _get_votes(self, L: MatrixLike) -> np.ndarray:
        """Compute the sum of votes for each label based on the given weak
        label matrix.

        Parameters
        ----------
        L : pd.DataFrame
            Weak label dataframe.

            Shape: (n_samples, n_weak)

        Returns
        -------
        votes : np.ndarray
            2-D array where votes[i.j] = sum votes for class j for sample i.

            Shape: (n_samples, n_classes)
        """
        L = np.array(L)
        votes = L @ self.polarities_matrix
        return votes

    def _normalize_preds(self, preds: np.ndarray) -> np.ndarray:
        # Normalize votes per row
        row_wise_sum = preds.sum(axis=1)
        # In case there were no votes for this row, no need to renormalize
        row_wise_sum = np.where(row_wise_sum == 0, 1, row_wise_sum).reshape(-1, 1)
        proba = preds / row_wise_sum
        return proba


class Voter(_BaseModel):
    """Basic model that bases its decisions on the sum of votes for each
    class."""

    def fit(self) -> None:
        """Voter model does not require fitting."""
        warnings.warn("Voter object does not need to be fitted", category=UserWarning)

    def predict_proba(self, L: MatrixLike) -> np.ndarray:
        """Predict probabilities using sum of votes.

        Parameters
        ----------
        L : pd.DataFrame
            Weak label dataframe.

        Returns
        -------
            Array of predicted probabilities of shape (len(L), cardinality)

        Example
        -------
        >>> L = [[1, 0, 1, 2], [0, 1, 0, 0]]
        >>> proba = snorkel_model.predict_proba(L)
        >>> # proba.shape = (len(L), cardinality)
        """
        votes = self._get_votes(L)
        proba = self._normalize_preds(votes)
        return proba


class BalancedVoter(_BaseModel):
    """Basic model that bases its decisions on a weighted sum of votes for each
    class.

    The weights are computed during fitting so the sum of votes over
    training matches the given class balance.
    """

    def fit(self, L: MatrixLike, class_balances: ListLike = []) -> None:
        """Fit the BalancedMajorityVoter model.

        Parameters
        ----------
        L
            Weak label dataframe.
        class_balances
            Numpy array of shape cardinality giving a weight to each class.

            By default, assumes all classes are equally likely.


        Example
        -------
        >>> L = [[1, 0, 1, 2], [0, 1, 2, 1], [1, 2, 1, 0], [0, 1, 0, 2]]
        >>> class_balances = [0.6, 0.4]
        >>> base_model.fit(L, class_balances)
        """
        if not class_balances:
            class_balances = np.ones(self.cardinality)
        class_balances = np.array(class_balances)
        self.class_balances = class_balances
        self.normalized_class_balances = self.class_balances / self.class_balances.sum()

        # Learn votes weights per class
        # vote weights is the factor that will reweigh the predicted probabilites

        # Compute the product between the weak label matrix and the polarities matrix
        # To get the sum of votes per label
        votes = self._get_votes(L)

        # Mean votes per class
        votes_mean = votes.mean(axis=0)
        # Deal with the case where there are no votes for a class
        no_votes = np.where(votes_mean == 0)[0]
        if no_votes.size > 0:
            for i in no_votes:
                warnings.warn(
                    (
                        f"No votes were given to labeling function {i},"
                        "assuming its balance is correct which is likely wrong"
                    ),
                    UserWarning,
                )
            votes_mean[no_votes] = self.normalized_class_balances[no_votes]
        votes_weights = self.normalized_class_balances / votes_mean
        self.votes_weights = votes_weights

    def predict_proba(self, L: MatrixLike) -> np.ndarray:
        """Predict probabilites using weighted voting.

        Parameters
        ----------
        L : pd.DataFrame
            Weak label dataframe.

        Returns
        -------
            Array of predicted probabilities of shape (len(L), cardinality)

        Example
        -------
        >>> L = [[1, 0, 1, 2], [0, 1, 0, 0]]
        >>> proba = snorkel_model.predict_proba(L)
        >>> # proba.shape = (len(L), cardinality)
        """
        votes = self._get_votes(L)
        weighted_votes = votes * self.votes_weights
        proba = self._normalize_preds(weighted_votes)
        return proba
